/*  
 *	Berkshire PC Watchdog Card Driver
 *
 *	Copyright (c) 1998-2002 Holger Eiboeck  <holger@eiboeck.de>, 
 *				  All Rights Reserved.
 *
 *	http://www.pcwd.de/
 *
 *	Idea and parts of the code by Kenji Hollis <khollis@bitgate.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	$Id: pcwd.c,v 1.42 2002/07/26 08:19:00 holgi Exp root $
 *
 *	$Log: pcwd.c,v $
 *	Revision 1.42  2002/07/26 08:19:00  holgi
 *	Added an extra spinlock variable for io operations only (David Chambers)
 *	to try to prevent deadlocks. Changed io operations to use this lock (David Chambers)
 *	Explicitly initialize spinlock_t pcwd_lock (David Chambers)
 *	indeted propperly for kernel submission
 *
 *	Revision 1.41  2002/06/11 12:47:47  holgi
 *	fixed dumb compile error when used direct in kernel found by Martin Josefson
 *	made CONFIG_WATCHDOG_NOWAYOUT a module parameter
 *	fixed a bug in case enabling/disabling the board failed
 *	full IOCTL support to be more compatible with other watchdog drivers
 *	reorganized some spinlocks
 *	separated possible /dev/watchdog and /dev/temperature operations
 *
 *	Revision 1.40  2001/12/18 10:28:17  holgi
 *	drop kernel < 2.4 support and all wrappers
 *	support pci hotplug
 *
 *  Revision 1.39  2001/12/18 10:22:09  holgi
 *  code reformat
 *  fixed bug in ISA proc output; now more reliable
 *  fixed Makefile for better 2.4 compatibility
 *  fixed isa card timeouts
 *  
 *	Revision 1.38  2001/09/05 12:38:07  holgi
 *	corrected bug in /proc buffer calculation
 *	new command 'tempoffset' (Olaf Christians)
 *
 *	Revision 1.37  2001/03/19 12:03:03  holgi
 *	forgot to initialize pcwd_lock (Martin Josefsson)
 *	more spin_locks (Friedrich Lobenstock)
 *	restore capability to compile directly into kernel
 *	tested compatibility with kernel 2.2.17 and 2.4.2
 *
 *	Revision 1.36  2001/01/23 18:30:06  holgi
 *	Friedrich Lobenstock <fl@fl.priv.at>:
 *	  code reformated to only use tabs to indent
 *	  new command 'debug', separate ISA & PCI timout
 *	  reset command has a 5 s delay
 *	  new commands 'clear', 'relay 1noinvert', 'relay 1nvnoinvert'
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/stat.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/reboot.h>
#include <linux/notifier.h>
#include <linux/init.h>
#include <linux/watchdog.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#ifndef PCI_VENDOR_ID_QUICKLOGIC
#define PCI_VENDOR_ID_QUICKLOGIC    0x11e3
#endif

#ifndef PCI_DEVICE_ID_WATCHDOG_PCWDPCI
#define PCI_DEVICE_ID_WATCHDOG_PCWDPCI 0x5030
#endif

static char rcsid[] = "$Revision: 1.42 $$Date: 2002/07/26 08:19:00 $";

/* revision */
#define	PCWD_REVISION_A		1
#define	PCWD_REVISION_C		2
#define PCWD_PCI			3
#define PCWD_USB			4

/* pcwd_temp_mode */
#define CELSIUS				0
#define FAHRENHEIT			1

/* pcwd_verbose - level of verbosity */
#define QUIET				0
#define VERBOSE				1	/* normal */
#define DEBUG				2	/* print fancy stuff too */

/* card_status, pcwd_diag_mode */
#define DISABLED		   	0
#define ENABLED				1

/* mode of operation for relay 2 */
#define TEMPERATURE_TRIP	0	/* activate relay 2 during a temperature trip */
#define NO_TEMPERATURE_TRIP	1	/* don't do it */

#define WD_TIMEOUT			3	/* 1 1/2 seconds for a timeout */

/* max. time we give an ISA watchdog card to process a command */
/* 500ms for each 4 bit response */
#define ISA_COMMAND_TIMEOUT	1000

/* according to documentaion max. time to process a command for the pci
   watchdog card is 100 ms, so we give it 150 ms to do it's job */
#define PCI_COMMAND_TIMEOUT	150

MODULE_PARM(nowayout, "i");
MODULE_PARM_DESC(nowayout,
		 "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif

static int revision = PCWD_REVISION_A;
static int supports_temp = 0;
static int in_use = 0;
static int card_status = DISABLED;
static int pcwd_verbose = VERBOSE;
// static int pcwd_verbose = DEBUG;
static int pcwd_temp_mode = CELSIUS;
static unsigned long card_ioport = 0x000;
static char pcwd_driver_version[40];
static char pcwd_fw_string[7];	/* firmware version xx.xxx */
static int pcwd_fw_rev_major, pcwd_fw_rev_minor;
static int pcwd_diag_mode = DISABLED;
static unsigned int ping_counter = 0;
static int initial_status, initial_status_set;
static int temp_mode_relay2 = TEMPERATURE_TRIP;
static int pcwd_temp_panic = DISABLED;
static spinlock_t pcwd_io_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pcwd_lock = SPIN_LOCK_UNLOCKED;

static void
pcwd_get_version(void)
{
	char *rcsvers, *rcsdate, *tmp;

	rcsvers = strchr(rcsid, ' ');
	rcsvers++;
	tmp = strchr(rcsvers, ' ');
	*tmp++ = '\0';
	rcsdate = strchr(tmp, ' ');
	rcsdate++;
	tmp = strrchr(rcsdate, ' ');
	*tmp = '\0';

	sprintf(pcwd_driver_version, "%s (%s)", rcsvers, rcsdate);
}

static int __init
pcwd_find_card(void)
{
	/* Available I/O port addresses for the PC Watchdog cards. */
	static int pcwd_ioports[] = { 0x270, 0x350, 0x370, 0 };

	int card_dat, prev_card_dat, count = 0, i;
	int rc = 0;
	int num_cards = 0;

	/* check for ISA cards */
	for (i = 0; pcwd_ioports[i] != 0; i++) {
		if (rc == 0) {
			card_dat = 0x00;
			if (check_region(pcwd_ioports[i], 4)) {
				printk(KERN_ERR
				       "pcwd: Port 0x%x unavailable.\n",
				       pcwd_ioports[i]);
				return 0;
			}

			spin_lock(&pcwd_io_lock);
			prev_card_dat = inb_p(pcwd_ioports[i]);

			count = 0;

			if (prev_card_dat != 0xFF) {
				while (count < WD_TIMEOUT) {
					card_dat = inb_p(pcwd_ioports[i] + 1);
					card_dat &= 0x02;
					mdelay(500);

					if (((card_dat == 0x02)
					     && (prev_card_dat == 0x00))
					    || ((card_dat == 0x00)
						&& (prev_card_dat == 0x02))) {
						/* heartbeat found */
						card_ioport = pcwd_ioports[i];
						num_cards++;
						rc = 1;
						break;
					}

					prev_card_dat = card_dat;
					count++;
				}
			}
			spin_unlock(&pcwd_io_lock);
		}
	}
	/* check for PCI cards */
	if (pci_present()) {
		struct pci_dev *dev = (0);
		while ((dev = pci_find_device(PCI_VENDOR_ID_QUICKLOGIC,
					      PCI_DEVICE_ID_WATCHDOG_PCWDPCI,
					      dev)) != (0)) {
			num_cards++;
			card_ioport = pci_resource_start(dev, 0);
			if (pci_enable_device(dev)) {
				printk(KERN_ERR
				       "pcwd: failed to enable watchdog card!\n");
				rc = 0;
			} else {
				revision = PCWD_PCI;
				rc = 1;
			}
		}
	}
	if (num_cards > 1) {
		printk(KERN_ERR "pcwd: more than 1 watchdog card found!\n");
		rc = 0;
	}

	return rc;
}

static int
pcwd_set_diag(void)
{
	if (revision == PCWD_PCI) {	/* PCI card has no diag mode */
		pcwd_diag_mode = ENABLED;

	} else {
		int i, found = 0, count = 0;

		while ((count < 3) && (!found)) {
			outb_p(0x80, card_ioport + 2);
			udelay(ISA_COMMAND_TIMEOUT);

			i = inb(card_ioport);

			if (i == 0x00)
				found = 1;
			else if (i == 0xF3)
				outb_p(0x00, card_ioport + 2);

			udelay(ISA_COMMAND_TIMEOUT);
			count++;
		}
		if (found) {
			pcwd_diag_mode = ENABLED;
		} else {
			pcwd_diag_mode = DISABLED;
		}
	}
	return pcwd_diag_mode;
}

static void
pcwd_diag_set_idle(void)
{
	if (revision != PCWD_PCI) {	/* PCI card has no diag mode -> no idle */
		outb_p(0x00, card_ioport + 2);
		udelay(ISA_COMMAND_TIMEOUT);

		(void) inb(card_ioport);
		(void) inb(card_ioport);
	}
}

/* send a command to watchdog card and give a 16 bit return value */
static int
send_command_ret16(int cmd, int msb, int lsb)
{
	static int response, count;

	/* send command to card */
	outb_p(lsb, card_ioport + 4);
	outb_p(msb, card_ioport + 5);
	outb_p(cmd, card_ioport + 6);

	if (revision == PCWD_PCI) {
		/* wait till the pci card processed the command, signaled by 
		   the WRSP bit in port 2 and give it a max. timeout of 
		   PCI_COMMAND_TIMEOUT to process 
		 */
		for (count = 0; (count < PCI_COMMAND_TIMEOUT)
		     && (!(inb(card_ioport + 2) & 0x40)); count++)
			mdelay(1);
		if (pcwd_verbose >= DEBUG)
			printk(KERN_INFO
			       "pcwd: time to process command was: %d ms\n",
			       count);

		/* read back response */
		response = inb(card_ioport + 4) + inb(card_ioport + 5) * 256;

		/* clear WRSP bit */
		inb(card_ioport + 6);

	} else {
		/* there is no status bit like the WRSP bit, so */
		/* wait max. timeout */
		mdelay(ISA_COMMAND_TIMEOUT);

		/* read back response */
		response = inb(card_ioport + 4) + inb(card_ioport + 5) * 256;
	}

	return response;
}

/* send a command to watchdog card and give a 8 bit return value */
static int
send_command_ret8(int cmd, int msb, int lsb)
{
	return (send_command_ret16(cmd, msb, lsb) & 0x00ff);
}

/* send a command to watchdog card and don't care for the result */
static inline void
send_command(int cmd, int msb, int lsb)
{
	send_command_ret16(cmd, msb, lsb);
}

static int
send_diag_command(int cmd)
{
	int i;

	if (revision == PCWD_PCI) {
		i = send_command_ret16(cmd, 0, 0);
	} else {
		if ((pcwd_diag_mode != ENABLED) && (pcwd_set_diag() != ENABLED)) {
			printk(KERN_ERR
			       "pcwd: send_diag_command diag mode failed");
			return (0);
		}
		outb_p(cmd, card_ioport + 2);
		udelay(ISA_COMMAND_TIMEOUT);

		i = inb(card_ioport);
		i = inb(card_ioport);
	}
	return i;
}

static int
pcwd_clear_status(void)
{
	if ((revision == PCWD_PCI) || (revision == PCWD_REVISION_C)) {
		int cdat;

		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: clearing trip status\n");

		cdat = inb_p(card_ioport + 1);

		if (pcwd_verbose >= DEBUG) {
			printk(KERN_DEBUG "pcwd: status was: 0x%02x\n", cdat);
			printk(KERN_DEBUG "pcwd: sending: 0x%02x\n",
			       (cdat & 0x40) | 0x01);
		}

		/* clear trip status & LED and keep mode of relay 2 */
		outb_p((cdat & 0x40) | 0x01, card_ioport + 1);

		/* clear reset counter */
		cdat = send_command_ret8(0x84, 0, 0xff);

		if (pcwd_verbose >= DEBUG) {
			printk(KERN_DEBUG "pcwd: reset count was: 0x%02x\n",
			       cdat);
		}

		/* set clear boot status */
		initial_status |= 0x02;

	} else {
		/* FIXME: does the ISA card revision A support this? */
		/* for now we suppose not */
		printk(KERN_ERR "pcwd: clear does not work on revision A\n");
		return 1;
	}
	return 0;
}

static int
pcwd_init_status(void)
{
	if ((revision == PCWD_PCI) || (revision == PCWD_REVISION_C)) {
		if (pcwd_verbose >= DEBUG)
			printk(KERN_INFO "pcwd: inititializing status\n");

		/* init status clear trip status & LED */
		outb_p((temp_mode_relay2) ? 0x40 : 0x00, card_ioport + 1);

	} else {
		/* FIXME: does a ISA card revision A support this? */
		/* for now we suppose not */
		return 1;
	}
	return 0;
}

static int
pcwd_reset_pc(void)
{
	if (revision == PCWD_PCI) {
		int rc;

		printk(KERN_INFO "pcwd: reseting PC!\n");

		/* reset with a 5 second delay - 
		 * user might want to umount, call sync, ... :) */

		rc = send_command_ret8(0x80, 0xa5, 5);

		/* FIXME: check ack from card, but it works... */

		card_status = DISABLED;
		return rc;

	} else {

		if (pcwd_set_diag() == ENABLED) {
			printk(KERN_INFO "pcwd: reseting PC!\n");

			send_diag_command(0x86);

			if (!(inb_p(card_ioport) == 0xAA)) {
				printk(KERN_ERR
				       "pcwd: Card did not acknowledge reset attempt.\n");
				return 0;
			}

			pcwd_diag_set_idle();
			printk(KERN_INFO "pcwd: reseting PC ACK!\n");

			card_status = DISABLED;
			return 1;
		}
	}
	return 0;
}

int
pcwd_disable_card(void)
{
	int stat_reg;

	spin_lock(&pcwd_io_lock);

	outb_p(0xA5, card_ioport + 3);
	udelay(ISA_COMMAND_TIMEOUT);

	outb_p(0xA5, card_ioport + 3);
	udelay(ISA_COMMAND_TIMEOUT);

	stat_reg = inb_p(card_ioport + 2);

	spin_unlock(&pcwd_io_lock);

	if (!(stat_reg & 0x10)) {
		printk(KERN_ERR
		       "pcwd: Card did not acknowledge disable attempt. Try again.\n");
		return 0;
	}

	card_status = DISABLED;

	return 1;
}

static int
pcwd_enable_card(void)
{
	int stat_reg;

	spin_lock(&pcwd_io_lock);

	outb_p(0x00, card_ioport + 3);
	udelay(ISA_COMMAND_TIMEOUT);
	stat_reg = inb_p(card_ioport + 2);

	spin_unlock(&pcwd_io_lock);

	if (stat_reg & 0x10) {
		printk(KERN_ERR "pcwd: Card timer not enabled.\n");
		return 0;
	}
	card_status = ENABLED;

	return 1;
}

static int
pcwd_get_temperature(void)
{
	int dat = 0;

	pcwd_diag_set_idle();

	if (supports_temp) {
		spin_lock(&pcwd_io_lock);
		dat = inb(card_ioport);
		dat = inb(card_ioport);
		spin_unlock(&pcwd_io_lock);
	} else {
		dat = 0;
	}

	if (pcwd_temp_mode == FAHRENHEIT)
		dat = (dat * 9 / 5) + 32;

	return dat;
}

static void
pcwd_get_relay2_mode(void)
{
	if (inb_p(card_ioport + 1) & 0x40)
		temp_mode_relay2 = NO_TEMPERATURE_TRIP;
	else
		temp_mode_relay2 = TEMPERATURE_TRIP;

	if (pcwd_verbose >= DEBUG)
		printk(KERN_INFO "pcwd: mode of relay 2 is: %s\n",
		       (temp_mode_relay2 ==
			TEMPERATURE_TRIP) ? "temp" : "notemp");
}

static int
pcwd_get_trip_status(void)
{
	int cdat;

	if (revision == PCWD_REVISION_A) {
		cdat = inb_p(card_ioport);
	} else {
		cdat = inb_p(card_ioport + 1);
	}

	if (initial_status_set == 0) {
		initial_status_set = 1;
		/* don't store HRBT bit - we need it to store the clear status */
		initial_status = cdat & (0xff - 0x02);
		if (pcwd_verbose >= DEBUG)
			printk(KERN_INFO "pcwd: initial status: 0x%02x\n",
			       initial_status);
	}

	return (cdat & 0x01);
}

static int
pcwd_get_switches(void)
{
	int i, j;

	if (revision == PCWD_PCI) {
		i = inb_p(card_ioport + 3);
		return i;

	} else {
		if (pcwd_set_diag() == ENABLED) {
			j = send_diag_command(0x85);
			udelay(ISA_COMMAND_TIMEOUT);
			i = send_diag_command(0x85);

			pcwd_diag_set_idle();
			return i;

		} else {
			return 0;
		}
	}
}

static inline void
get_support(void)
{
	if (inb(card_ioport) != 0xF0)
		supports_temp = 1;
	else
		supports_temp = 0;
}

static inline void
get_revision(void)
{
	if (revision != PCWD_PCI) {
		if ((inb(card_ioport + 2) == 0xFF)
		    || (inb(card_ioport + 3) == 0xFF))
			revision = PCWD_REVISION_A;
		else
			revision = PCWD_REVISION_C;
	}
}

static void
pcwd_get_fw_string(void)
{
	if (revision == PCWD_PCI) {
		int firmware_revision;

		firmware_revision = send_command_ret16(0x08, 0, 0);
		pcwd_fw_rev_major = firmware_revision / 256;
		pcwd_fw_rev_minor = firmware_revision % 256;

		sprintf(pcwd_fw_string, "%u.%02u", pcwd_fw_rev_major,
			pcwd_fw_rev_minor);

	} else {
		if (pcwd_set_diag() == ENABLED) {
			int one, ten, hund, minor;

			one = send_diag_command(0x81);
			ten = send_diag_command(0x82);
			hund = send_diag_command(0x83);
			minor = send_diag_command(0x84);
			sprintf(pcwd_fw_string, "%c.%c%c%c", one, ten, hund,
				minor);
		} else {
			sprintf(pcwd_fw_string, "ERROR");
		}
	}
}

static void
pcwd_show_card_stats(void)
{
	switch (revision) {
	case PCWD_REVISION_A:
		printk(KERN_INFO
		       "pcwd: PC Watchdog Card at 0x%x: REV.A %s temp option.\n",
		       (int) card_ioport, (supports_temp ? "with" : "without"));
		break;

	case PCWD_REVISION_C:
		printk(KERN_INFO
		       "pcwd: PC Watchdog Card at 0x%x: REV.C (Firmware: %s) %s temp option.\n",
		       (int) card_ioport, pcwd_fw_string,
		       (supports_temp ? "with" : "without"));
		break;

	case PCWD_PCI:
		printk(KERN_INFO
		       "pcwd: PCI PC Watchdog Card at 0x%x: (Firmware: %s) %s temp option.\n",
		       (int) card_ioport, pcwd_fw_string,
		       (supports_temp ? "with" : "without"));
		break;

	default:
		printk(KERN_ERR
		       "pcwd: Unknown or unsupported card or revision.\n");
	}

	pcwd_get_trip_status();

	if (initial_status & 0x01)
		printk(KERN_INFO
		       "pcwd: Previous reset was caused by the Watchdog card.\n");
	if (initial_status & 0x04)
		printk(KERN_INFO "pcwd: Card sensed a CPU Overheat.\n");
	if (!(initial_status & 0x01))
		printk(KERN_INFO
		       "pcwd: No previous trip detected - Cold boot or reset.\n");
}

static void
pcwd_ping(void)
{
	if (revision == PCWD_PCI) {
		spin_lock(&pcwd_io_lock);
		outb_p(0xfc, card_ioport);	/* send out any data */
		spin_unlock(&pcwd_io_lock);
	} else {

		int wdrst_stat;

		spin_lock(&pcwd_io_lock);

		wdrst_stat = inb_p(card_ioport);
		wdrst_stat &= 0x0F;
		wdrst_stat |= 0x01;	/* Previously reset state */

		if (revision == PCWD_REVISION_A)
			outb_p(wdrst_stat, card_ioport + 1);
		else
			outb_p(wdrst_stat, card_ioport);

		spin_unlock(&pcwd_io_lock);
	}
	if (card_status == ENABLED)
		ping_counter++;
}

/* The file operation functions for user access to /dev/watchdog */
static int
pcwd_open(struct inode *inode, struct file *file)
{
	switch (MINOR(inode->i_rdev)) {
	case WATCHDOG_MINOR:
		spin_lock(&pcwd_lock);
		if (in_use) {
			spin_unlock(&pcwd_lock);
			printk(KERN_ERR
			       "pcwd: Attempt to open already opened device.\n");
			return -EBUSY;
		}

		MOD_INC_USE_COUNT;

		in_use++;
		spin_unlock(&pcwd_lock);
		return 0;

	case TEMP_MINOR:
		MOD_INC_USE_COUNT;
		return 0;

	default:
		return -ENODEV;
	}
}

static int
pcwd_release(struct inode *inode, struct file *file)
{
	lock_kernel();
	MOD_DEC_USE_COUNT;
	if (MINOR(inode->i_rdev) == WATCHDOG_MINOR) {
		spin_lock(&pcwd_lock);
		if (!nowayout)
			pcwd_disable_card();
		in_use--;
		spin_unlock(&pcwd_lock);
	}
	unlock_kernel();
	return 0;
}

/*
static long long
pcwd_llseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}
*/

static ssize_t
pcwd_read(struct file *file, char *buf, size_t len, loff_t * ppos)
{
	unsigned short temp;

	temp = pcwd_get_temperature();

	/*  Can't seek (pread) on this device */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	switch (MINOR(file->f_dentry->d_inode->i_rdev)) {
	case TEMP_MINOR:
	case WATCHDOG_MINOR:
		if (copy_to_user(buf, &temp, 1))
			return -EFAULT;
		return 0;

	default:
		return -EINVAL;
	}
}

static ssize_t
pcwd_write(struct file *file, const char *buf, size_t len, loff_t * ppos)
{
	/*  Can't seek (pwrite) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (len) {
		pcwd_ping();
		return 1;
	}
	return 0;
}

static int
pcwd_ioctl(struct inode *inode, struct file *file,
	   unsigned int cmd, unsigned long arg)
{
	int rv;
	int retval;

	static struct watchdog_info ident = {
		WDIOF_OVERHEAT | WDIOF_CARDRESET,
		1,
		"PCWD"
	};

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		rv = copy_to_user((void *) arg, &ident, sizeof (ident));
		return rv ? -EFAULT : 0;

	case WDIOC_GETSTATUS:
		if (initial_status & WDIOF_OVERHEAT) {
			if (pcwd_temp_panic == ENABLED)
				panic("pcwd: Temperature overheat trip!\n");
		}

		if (put_user(initial_status, (int *) arg))
			return -EFAULT;
		return 0;

	case WDIOC_GETBOOTSTATUS:
		if (put_user(initial_status, (int *) arg))
			return -EFAULT;
		return 0;

	case WDIOC_GETTEMP:
		rv = 0;
		if (supports_temp)
			rv = pcwd_get_temperature();

		if (put_user(rv, (int *) arg))
			return -EFAULT;
		return 0;

	case WDIOC_SETOPTIONS:
		if (copy_from_user(&rv, (int *) arg, sizeof (int)))
			return -EFAULT;

		retval = -EINVAL;

		if (rv & WDIOS_DISABLECARD) {
			if (!pcwd_disable_card())
				return -EIO;
			retval = 0;
		}

		if (rv & WDIOS_ENABLECARD) {
			if (!pcwd_enable_card())
				return -EIO;
			retval = 0;
		}

		if (rv & WDIOS_TEMPPANIC) {
			pcwd_temp_panic = ENABLED;
			retval = 0;
		}

		return retval;

	case WDIOC_KEEPALIVE:
		pcwd_ping();
		return 0;

	default:
		return -ENOTTY;
	}
}

#ifdef CONFIG_PROC_FS
static int
pcwd_set_relay(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("relay value: %d\n", send_command_ret8(0x50, 0, 0));
		return 0;
	}

	i = send_command_ret8(0x50, 0, 0);	// read the relays

	if (strcmp(sub_command, "1on") == 0) {
		i = send_command_ret8(0x51, 0, (i | 1) & (0xff - 4));
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 1on\n");
	} else if (strcmp(sub_command, "1off") == 0) {
		i = send_command_ret8(0x51, 0, (i | 4) & (0xff - 1));
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 1off\n");

	} else if (strcmp(sub_command, "2on") == 0) {
		i = send_command_ret8(0x51, 0, (i | 2) & (0xff - 8));
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 2on\n");

	} else if (strcmp(sub_command, "2off") == 0) {
		i = send_command_ret8(0x51, 0, (i | 8) & (0xff - 2));
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 2off\n");

	} else if (strcmp(sub_command, "1invert") == 0) {
		if (pcwd_verbose >= DEBUG) {
			printk(KERN_DEBUG "pcwd: before: 0x%02x\n", i);
			printk(KERN_DEBUG "pcwd: sending: 0x%02x\n", i | 0x10);
		}
		i = send_command_ret8(0x51, 0, i | 0x10);
		if (pcwd_verbose >= DEBUG) {
			i = send_command_ret8(0x50, 0, 0);	// read the relays
			printk(KERN_DEBUG "pcwd: after: 0x%02x\n", i);
		}
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 1invert\n");

	} else if (strcmp(sub_command, "1noinvert") == 0) {
		if (pcwd_verbose >= DEBUG) {
			printk(KERN_DEBUG "pcwd: before: 0x%02x\n", i);
			printk(KERN_DEBUG "pcwd: sending: 0x%02x\n",
			       i & (0xff - 0x10));
		}
		i = send_command_ret8(0x51, 0, i & (0xff - 0x10));
		if (pcwd_verbose >= DEBUG) {
			i = send_command_ret8(0x50, 0, 0);	// read the relays
			printk(KERN_DEBUG "pcwd: after: 0x%2x\n", i);
		}
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 1noinvert\n");

	} else if (strcmp(sub_command, "1nvinvert") == 0) {
		if (pcwd_verbose >= DEBUG) {
			printk(KERN_DEBUG "pcwd: before: 0x%02x\n", i);
			printk(KERN_DEBUG "pcwd: sending: 0x%02x\n", i | 0x80);
		}
		i = send_command_ret8(0x51, 0, i | 0x80);
		if (pcwd_verbose >= DEBUG) {
			i = send_command_ret8(0x50, 0, 0);	// read the relays
			printk(KERN_DEBUG "pcwd: after: 0x%02x\n", i);
		}
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 1nvinvert\n");

	} else if (strcmp(sub_command, "1nvnoinvert") == 0) {
		if (pcwd_verbose >= DEBUG) {
			printk(KERN_DEBUG "pcwd: before: 0x%02x\n", i);
			printk(KERN_DEBUG "pcwd: sending: 0x%02x\n",
			       i & (0xff - 0x80));
		}
		i = send_command_ret8(0x51, 0, i & (0xff - 0x80));
		if (pcwd_verbose >= DEBUG) {
			i = send_command_ret8(0x50, 0, 0);	// read the relays
			printk(KERN_DEBUG "pcwd: after: 0x%02x\n", i);
		}
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 1nvnoinvert\n");

	} else if (strcmp(sub_command, "2notemp") == 0) {
		outb_p(0x40, card_ioport + 1);
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 2notemp\n");
		pcwd_get_relay2_mode();

	} else if (strcmp(sub_command, "2temp") == 0) {
		outb_p(0x00, card_ioport + 1);
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: relay 2temp\n");
		pcwd_get_relay2_mode();

	} else {
		printk(KERN_ERR
		       "pcwd: illegal relay argument '%s', should be 1on, 1off, 2on, 2off, 1invert, 1noinvert, 1nvinvert, 1nvnoinvert, 2temp or 2notemp\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcwd_set_dio(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("dio value: %d\n", (inb(card_ioport + 7) / 16) & 15);
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i < 16)) {
		outb_p(i, card_ioport + 7);
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: dio %d\n", i);
	} else {
		printk(KERN_ERR
		       "pcwd: illegal dio argument '%s', should be a number between 0 and 15\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcwd_set_nv_timeout_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("nvtimeout value: %d\n", send_command_ret16(0x1C, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if (i == 0)
		printk("returning to default DIP switch settings.\n");

	if ((i >= 0) && (i <= 0xffff)) {
		send_command(0x1D, i / 256, i % 256);
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: nvtimeout %d\n", i);
	} else {
		printk(KERN_ERR
		       "pcwd: illegal nvtimeout argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcwd_set_timeout_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("timeout value: %d\n", send_command_ret16(0x18, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i <= 0xffff)) {
		send_command(0x19, i / 256, i % 256);
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: timeout %d\n", i);
	} else {
		printk(KERN_ERR
		       "pcwd: illegal timeout argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcwd_set_timeout(char *sub_command)
{
	char *new_sub_command = sub_command;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk(KERN_ERR
		       "pcwd: illegal timeout argument '%s', should be 2, 4 or 8\n",
		       sub_command);
		return 1;
	}

	/* FIXME: show new timeouts also in proc */

	if (strcmp(sub_command, "2") == 0) {
		if (send_diag_command(0x8A) == 0x8A) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: timeout 2s\n");
			return 0;
		}
	} else if (strcmp(sub_command, "4") == 0) {
		if (send_diag_command(0x8B) == 0x8b) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: timeout 4s\n");
			return 0;
		}
	} else if (strcmp(sub_command, "8") == 0) {
		if (send_diag_command(0x8C) == 0x8C) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: timeout 8s\n");
			return 0;
		}
	} else {
		printk(KERN_ERR
		       "pcwd: illegal timeout argument '%s', should be 2, 4 or 8\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcwd_set_nv_arm_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("nvarm value: %d\n", send_command_ret16(0x14, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i <= 0xffff)) {
		send_command(0x15, i / 256, i % 256);
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: nvarm %d\n", i);
	} else {
		printk(KERN_ERR
		       "pcwd: illegal nvarm argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcwd_set_arm_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("arm value: %d\n", send_command_ret16(0x10, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i <= 0xffff)) {
		if (send_command_ret8(0x11, i / 256, i % 256) == 0) {
			printk(KERN_INFO
			       "pcwd: arm failed, board already armed!\n");
			return 1;
		}
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: arm %d\n", i);
	} else {
		printk(KERN_ERR
		       "pcwd: illegal arm argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcwd_set_arm(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i = 0;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk(KERN_ERR
		       "pcwd: illegal arm argument '%s', should be 0, 30 or 60\n",
		       sub_command);
		return 1;
	}

	pcwd_set_diag();

	if ((strcmp(sub_command, "0") == 0)
	    || (strcmp(sub_command, "immediately") == 0)) {
		if ((i = send_diag_command(0x87)) == 0x87) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: arm immediately\n");
			return 0;
		}
	} else if (strcmp(sub_command, "30") == 0) {
		if ((i = send_diag_command(0x88)) == 0x88) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: arm after 30s more\n");
			return 0;
		}
	} else if (strcmp(sub_command, "60") == 0) {
		if ((i = send_diag_command(0x89)) != 0x89) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: arm after 60s more\n");
			return 0;
		}
	} else {
		printk(KERN_ERR
		       "pcwd: illegal arm argument '%s', should be 0, 30 or 60\n",
		       sub_command);
		return 1;
	}

	if (i == 0xF5)
		printk(KERN_INFO "pcwd: already armed\n");

	return 0;
}

static int
pcwd_set_tempoffset_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("tempoffset value: %d\n", (inb(card_ioport + 2) & 15));
		return 0;
	}

	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i < 16)) {
		outb_p(i, card_ioport + 2);
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: tempoffset %d\n", i);
	} else {
		printk(KERN_ERR
		       "pcwd: illegal tempoffset argument '%s', should be a number between 0 and 15\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static void
pcwd_user_help(void)
{
	printk(" pcwd proc command interface help:\n");
	printk("   ping, pong, tickle - tickle the timer\n");
	printk("   hardreset - reset pc\n");
	printk("   disable, enable - set watchdog mode\n");
	printk("   clear - clear trip status and led - pci card only\n");
	printk("   verbose, debug, quiet - command reponse\n");
	printk("   fahrenheit, celsius - display temp in\n");
	printk(" extended commands:\n");
	printk("   arm [val] - isa [30,60,90], pci [any]\n");
	printk("   nvarm [val] - pci card only\n");
	printk("   timeout [val] - isa [2,4,8], pci [any]\n");
	printk("   nvtimeout [val] - pci card only\n");
	printk("   dio [val] - pci card only\n");
	printk("   relay [1on,1off,2on,2off,1invert,1noinvert,\n");
	printk
	    ("          1nvinvert,1nvnoinvert,2temp,2notemp] - pci card only\n");
	printk("    tempoffset [val] - pci card only\n");
}

static int
pcwd_user_command(char *user_command)
{
	if ((strcmp(user_command, "ping") == 0)
	    || (strcmp(user_command, "pong") == 0)
	    || (strcmp(user_command, "tickle") == 0)) {
		pcwd_ping();
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: ping...\n");
	} else if (strcmp(user_command, "disable") == 0) {
		if (pcwd_disable_card()) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: watchdog disabled!\n");
		} else {
			printk(KERN_ERR "pcwd: disable watchdog FAILED!\n");
		}
	} else if (strcmp(user_command, "enable") == 0) {
		if (pcwd_enable_card()) {
			if (pcwd_verbose >= VERBOSE)
				printk(KERN_INFO "pcwd: watchdog enabled.\n");
		} else {
			printk(KERN_ERR "pcwd: enable watchdog FAILED.\n");
		}
	} else if (strcmp(user_command, "clear") == 0) {
		pcwd_clear_status();
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: cleared trip status\n");
	} else if (strcmp(user_command, "hardreset") == 0) {
		pcwd_reset_pc();
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: hardreset PC!\n");
	} else if (strcmp(user_command, "verbose") == 0) {
		printk(KERN_INFO "pcwd: verbose mode on.\n");
		pcwd_verbose = VERBOSE;
	} else if (strcmp(user_command, "debug") == 0) {
		printk(KERN_INFO "pcwd: debug mode on.\n");
		pcwd_verbose = DEBUG;
	} else if (strcmp(user_command, "quiet") == 0) {
		pcwd_verbose = QUIET;
	} else if (strcmp(user_command, "celsius") == 0) {
		pcwd_temp_mode = CELSIUS;
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: display temp in Celsius.\n");
	} else if (strcmp(user_command, "fahrenheit") == 0) {
		pcwd_temp_mode = FAHRENHEIT;
		if (pcwd_verbose >= VERBOSE)
			printk(KERN_INFO "pcwd: display temp in Fahrenheit.\n");
	} else if (strcmp(user_command, "help") == 0) {
		pcwd_user_help();
	} else if (strncmp(user_command, "arm", 3) == 0) {
		if (revision == PCWD_PCI) {
			if (pcwd_set_arm_pci(&user_command[3])) {
				printk(KERN_ERR "arm FAILED.\n");
				return 0;
			}
		} else {	// ISA CARD
			if ((pcwd_fw_string[0] >= '1')
			    && (pcwd_fw_string[2] >= '2')) {
				if (pcwd_set_arm(&user_command[3])) {
					printk(KERN_ERR "arm FAILED.\n");
					return 0;
				}
			} else {
				printk(KERN_ERR
				       "pcwd: ISA Card and firmware > 1.20 needed.\n");
				return 0;
			}
		}
	} else if (strncmp(user_command, "nvarm", 5) == 0) {
		if (revision == PCWD_PCI) {
			if (pcwd_set_nv_arm_pci(&user_command[5])) {
				printk(KERN_ERR "nvarm FAILED.\n");
				return 0;
			}
		} else {	// ISA CARD
			printk(KERN_ERR
			       "pcwd: nvarm only works with a pci card.\n");
			return 0;
		}
	} else if (strncmp(user_command, "timeout", 7) == 0) {
		if (revision == PCWD_PCI) {
			if (pcwd_set_timeout_pci(&user_command[7])) {
				printk(KERN_ERR "timeout FAILED.\n");
				return 0;
			}
		} else {	// ISA CARD
			if ((pcwd_fw_string[0] >= '1')
			    && (pcwd_fw_string[2] >= '2')) {
				if (pcwd_set_timeout(&user_command[7])) {
					printk(KERN_ERR "timeout FAILED.\n");
					return 0;
				}
			} else {	// ISA FIRMWARE TOO OLD
				printk(KERN_ERR
				       "pcwd: ISA Card and firmware > 1.20 needed.\n");
				return 0;
			}
		}
	} else if (strncmp(user_command, "nvtimeout", 9) == 0) {
		if (revision == PCWD_PCI) {
			if (pcwd_set_nv_timeout_pci(&user_command[9])) {
				printk(KERN_ERR "nvtimeout FAILED.\n");
				return 0;
			}
		} else {	// ISA CARD
			printk(KERN_ERR
			       "pcwd: nvtimeout only works with a pci card.\n");
			return 0;
		}
	} else if (strncmp(user_command, "dio", 3) == 0) {
		if (revision == PCWD_PCI) {
			if (pcwd_set_dio(&user_command[3])) {
				printk(KERN_ERR "dio FAILED.\n");
				return 0;
			}
		} else {	// ISA CARD
			printk(KERN_ERR
			       "pcwd: dio only works with a pci card.\n");
			return 0;
		}
	} else if (strncmp(user_command, "relay", 5) == 0) {
		if (revision == PCWD_PCI) {
			if (pcwd_set_relay(&user_command[5])) {
				printk(KERN_ERR "relay FAILED.\n");
				return 0;
			}
		} else {	// ISA CARD
			printk(KERN_ERR
			       "pcwd: relay only works with a pci card.\n");
			return 0;
		}
	} else if (strncmp(user_command, "tempoffset", 10) == 0) {
		if (revision == PCWD_PCI) {
			if (pcwd_set_tempoffset_pci(&user_command[10])) {
				printk(KERN_ERR "tempoffset FAILED.\n");
				return 0;
			}
		} else {	// ISA CARD
			printk(KERN_ERR
			       "pcwd: tempoffset only works with a pci card.\n");
			return 0;
		}
	} else {
		printk(KERN_ERR "illegal user command: '%s'\n", user_command);
		pcwd_user_help();
		return 0;
	}
	return 1;
}

static int
pcwd_write_proc(struct file *file, const char *buffer,
		unsigned long count, void *data)
{
	char command_buffer[80];
	int rc, length;

	/* write only allowed for root users */
	if (current->euid != 0)
		return -EACCES;

	if (count > sizeof (command_buffer) - 1)
		return -EINVAL;

	if (copy_from_user(command_buffer, buffer, count))
		return -EFAULT;

	command_buffer[count] = '\0';
	length = strlen(command_buffer);
	if (command_buffer[length - 1] == '\n')
		command_buffer[--length] = '\0';

	spin_lock(&pcwd_lock);
	rc = pcwd_user_command(command_buffer);
	spin_unlock(&pcwd_lock);
	return (rc ? count : -EBUSY);
}

static int pcwd_proc_info(unsigned char *contents, char *buffer,
			  int *len, off_t * begin, off_t offset, int size);

static struct proc_dir_entry *proc_pcwd;

static int
pcwd_read_proc(char *buffer, char **start, off_t offset,
	       int size, int *eof, void *data)
{
	int len = 0;
	off_t begin = 0;

	spin_lock(&pcwd_lock);
	*eof = pcwd_proc_info(NULL, buffer, &len, &begin, offset, size);
	spin_unlock(&pcwd_lock);

	if (offset >= begin + len)
		return (0);
	*start = buffer + (offset - begin);	// CORRECTED !!!!
	return (size < begin + len - offset ? size : begin + len - offset);
}

/* This macro frees the machine specific function from bounds checking and
 *  * this like that... [from nvram.c]*/
#define PRINT_PROC(fmt,args...)                         \
    do {                                                \
        *len += sprintf( buffer+*len, fmt, ##args );    \
        if (*begin + *len > offset + size)              \
            return( 0 );                                \
        if (*begin + *len < offset) {                   \
            *begin += *len;                             \
            *len = 0;                                   \
        }                                               \
    } while(0)

static int
pcwd_proc_info(unsigned char *buf, char *buffer, int *len,
	       off_t * begin, off_t offset, int size)
{
	int sw, i, j, t = 0;

	if (revision == PCWD_REVISION_C) {
		PRINT_PROC("PC Watchdog Driver %s\n", pcwd_driver_version);
		PRINT_PROC("Firmware     : Version %s\n", pcwd_fw_string);
		PRINT_PROC("Option Switch: Delay Time ");
		sw = pcwd_get_switches();
		switch (sw & 0x07) {
		case 0:
			PRINT_PROC("20 s");
			break;
		case 1:
			PRINT_PROC("40 s");
			break;
		case 2:
			PRINT_PROC("1 min");
			break;
		case 3:
			PRINT_PROC("5 min");
			break;
		case 4:
			PRINT_PROC("10 min");
			break;
		case 5:
			PRINT_PROC("30 min");
			break;
		case 6:
			PRINT_PROC("1 h");
			break;
		case 7:
			PRINT_PROC("2 h");
			break;
		}
		if (sw & 0x08) {
			PRINT_PROC(", POD on");
		} else {
			PRINT_PROC(", POD off");
		}
		if (sw & 0x10) {
			PRINT_PROC(", TRE on");
		} else {
			PRINT_PROC(", TRE off");
		}
		if (sw & 0x20) {
			PRINT_PROC(", R2M on");
		} else {
			PRINT_PROC(", R2M off");
		}
		if (sw & 0x40) {
			PRINT_PROC(", R1M on");
		} else {
			PRINT_PROC(", R1M off");
		}
		if (sw & 0x80) {
			PRINT_PROC(", RTM on\n");
		} else {
			PRINT_PROC(", RTM off\n");
		}

	} else if (revision == PCWD_PCI) {
		PRINT_PROC("PC PCI Watchdog Driver %s\n", pcwd_driver_version);
		PRINT_PROC("Firmware     : Version %s\n", pcwd_fw_string);
		PRINT_PROC("Option Switch: Delay Time ");
		sw = pcwd_get_switches();
		switch (sw & 0x07) {
		case 0:
			PRINT_PROC("5 s");
			t = 5;
			break;
		case 1:
			PRINT_PROC("10 s");
			t = 10;
			break;
		case 2:
			PRINT_PROC("30 s");
			t = 30;
			break;
		case 3:
			PRINT_PROC("1 min");
			t = 60;
			break;
		case 4:
			PRINT_PROC("5 min");
			t = 300;
			break;
		case 5:
			PRINT_PROC("10 min");
			t = 600;
			break;
		case 6:
			PRINT_PROC("30 min");
			t = 1800;
			break;
		case 7:
			PRINT_PROC("1 h");
			t = 3600;
			break;
		}
		if (sw & 0x08) {
			PRINT_PROC(", POD on");
		} else {
			PRINT_PROC(", POD off");
		}
		if (sw & 0x10) {
			PRINT_PROC(", TRE on");
		} else {
			PRINT_PROC(", TRE off");
		}
		if (sw & 0x20) {
			PRINT_PROC(", R2M on");
		} else {
			PRINT_PROC(", R2M off");
		}
		if (sw & 0x40) {
			PRINT_PROC(", R1M on");
		} else {
			PRINT_PROC(", R1M off");
		}
		if (sw & 0x80) {
			PRINT_PROC(", RTM on\n");
		} else {
			PRINT_PROC(", RTM off\n");
		}

		sw = send_command_ret8(0x50, 0, 0);

		PRINT_PROC("Relays       : ");
		if (sw & 0x01) {
			PRINT_PROC("Relay 1 on");
		} else {
			PRINT_PROC("Relay 1 off");
		}
		if (sw & 0x02) {
			PRINT_PROC(", Relay 2 on");
		} else {
			PRINT_PROC(", Relay 2 off");
		}
		PRINT_PROC("\n               ");

		if (sw & 0x10) {
			PRINT_PROC("Relay 1 invert, ");
		}
		if (sw & 0x80) {
			PRINT_PROC("Relay 1 invert in nvram, ");
		}
		if (temp_mode_relay2 == TEMPERATURE_TRIP) {
			PRINT_PROC("Relay 2 temp");
		} else {
			PRINT_PROC("Relay 2 notemp");
		}

		PRINT_PROC("\n");

		i = send_command_ret16(0x14, 0, 0);	// ARM NVRAM
		j = send_command_ret16(0x10, 0, 0);	// ARM current
		PRINT_PROC("Arm Time     : %d s (NVRAM: %d s)\n", j, i);

		i = send_command_ret16(0x1C, 0, 0);	// Time NVRAM
		j = send_command_ret16(0x18, 0, 0);	// Time current
		PRINT_PROC("Timer        : %d s (NVRAM: %d s)\n", j, i);

		PRINT_PROC("Status       : ");
		sw = send_command_ret8(0x04, 0, 0);
		if (sw & 0x01) {
			PRINT_PROC("NVRAM OK");
		} else {
			PRINT_PROC("NVRAM FAILED");
		}
		if (sw & 0x02) {
			PRINT_PROC(", TEMP SENSOR OK");
		} else {
			PRINT_PROC(", TEMP SENSOR FAILED");
		}
		if (sw & 0x80) {
			PRINT_PROC(", POD COMPLETED");
		} else {
			PRINT_PROC(", POD ACTIVE");
		}
		PRINT_PROC(".\n");

		sw = inb(card_ioport + 7);
		PRINT_PROC("Digi I/O Port: OUT=%d, IN=%d\n", (sw & 15),
			   ((sw / 16) & 15));

		sw = send_command_ret8(0x84, 0, 0);
		PRINT_PROC("Reset Count  : %d\n", sw);

		sw = inb(card_ioport + 2) & 15;
		PRINT_PROC("Temp.TripPt. : LOWER=%d C UPPER=%d C\n", 46 + sw,
			   56 + sw);

	} else {
		PRINT_PROC("Firmware ROM not present\n");
	}

	PRINT_PROC("Temperature  : ");
	if (supports_temp) {
		PRINT_PROC("%d ", pcwd_get_temperature());
		if (pcwd_temp_mode == CELSIUS)
			PRINT_PROC("C\n");
		else
			PRINT_PROC("F\n");
	} else {
		PRINT_PROC("Card without temp option\n");
	}

	if (card_status == ENABLED)
		PRINT_PROC("Card status  : Timer enabled\n");
	else
		PRINT_PROC("Card status  : Timer disabled\n");

	if (pcwd_get_trip_status())
		PRINT_PROC("Reboot status: Tripped\n");
	else
		PRINT_PROC("Reboot status: Not tripped\n");

	if (pcwd_verbose >= DEBUG)
		printk(KERN_INFO "pcwd: initial_status: 0x%02x\n",
		       initial_status);

	if (initial_status & 0x02) {
		PRINT_PROC("Boot status  : cleared by operator\n");
	} else if ((initial_status & 0x01) && (initial_status & 0x04)) {
		PRINT_PROC
		    ("Boot status  : CPU Overheat  + Watchdog caused reboot\n");
	} else if (initial_status & 0x01) {
		PRINT_PROC("Boot status  : Watchdog caused reboot\n");
	} else if (initial_status & 0x04) {
		PRINT_PROC("Boot status  : CPU Overheat\n");
	} else {
		PRINT_PROC("Boot status  : Cold boot or reset\n");
	}

	// PRINT_PROC("Ping Count   : %d\n", ping_counter);
	return 1;
}
#endif				/* PROC_FS */

static int
pcwd_notify_sys(struct notifier_block *this, unsigned long code, void *unused)
{
	if (code == SYS_DOWN || code == SYS_HALT) {
		pcwd_disable_card();
		printk(KERN_INFO
		       "pcwd: reboot notifier has disabled the card!\n");
		//printk(KERN_INFO "pcwd: reboot notifier did not disable the card!\n");
	}
	return NOTIFY_DONE;
}

static struct file_operations pcwd_fops = {
	owner:THIS_MODULE,
	write:pcwd_write,
	ioctl:pcwd_ioctl,
	open:pcwd_open,
	release:pcwd_release,
};

static struct miscdevice pcwd_miscdev = {
	WATCHDOG_MINOR,
	"watchdog",
	&pcwd_fops
};

static struct file_operations pcwd_temp_fops = {
	owner:THIS_MODULE,
	read:pcwd_read,
	open:pcwd_open,
	release:pcwd_release,
};

static struct miscdevice temp_miscdev = {
	TEMP_MINOR,
	"temperature",
	&pcwd_temp_fops
};

static struct notifier_block pcwd_notifier = {
	pcwd_notify_sys,
	NULL,
	0
};

int __init
pcwatchdog_init(void)
{
	spin_lock_init(&pcwd_io_lock);
	spin_lock_init(&pcwd_lock);

	pcwd_get_version();

	printk(KERN_INFO "pcwd: %s by <holger@eiboeck.de>\n",
	       pcwd_driver_version);

	initial_status = 0;
	initial_status_set = 0;

	if ((!pcwd_find_card()) || (card_ioport == 0x000)) {
		printk(KERN_ERR "pcwd: No card detected.\n");
		return -ENODEV;
	}

	pcwd_disable_card();

	get_support();
	get_revision();
	pcwd_get_fw_string();

	switch (revision) {
	case PCWD_REVISION_C:
		request_region(card_ioport, 4, "PCWD Rev C");
		break;
	case PCWD_REVISION_A:
		request_region(card_ioport, 2, "PCWD Rev A");
		break;
	case PCWD_PCI:
		request_region(card_ioport, 8, "PCWD PCI");
		break;
	default:
		printk(KERN_ERR "pcwd: Unknown card revision.\n");
		return -ENODEV;
		break;
	}

	pcwd_show_card_stats();
	pcwd_init_status();

	if (misc_register(&pcwd_miscdev) != 0) {
		release_region(card_ioport, (revision == PCWD_REVISION_A) ?
			       2 : (revision == PCWD_REVISION_C) ? 4 : 8);
		return -EIO;
	}
	if (supports_temp)
		if (misc_register(&temp_miscdev) != 0) {
			misc_deregister(&pcwd_miscdev);
			release_region(card_ioport,
				       (revision ==
					PCWD_REVISION_A) ? 2 : (revision ==
								PCWD_REVISION_C)
				       ? 4 : 8);
			return -EIO;
		}
#ifdef CONFIG_PROC_FS
	if ((proc_pcwd =
	     create_proc_entry("pcwd", S_IFREG | S_IRUGO | S_IWUSR, 0))) {
		proc_pcwd->read_proc = pcwd_read_proc;
		proc_pcwd->write_proc = pcwd_write_proc;
	}
#endif

	if (nowayout)
		printk(KERN_INFO
		       "pcwd: Watchdog cannot be stopped once started\n");

	register_reboot_notifier(&pcwd_notifier);

	return 0;
}

void __exit
pcwatchdog_exit(void)
{

#ifdef CONFIG_PROC_FS
	if (proc_pcwd)
		remove_proc_entry("pcwd", 0);
#endif
	lock_kernel();

	if (!nowayout)
		pcwd_disable_card();

	misc_deregister(&pcwd_miscdev);
	if (supports_temp)
		misc_deregister(&temp_miscdev);

	unregister_reboot_notifier(&pcwd_notifier);

	release_region(card_ioport,
		       (revision == PCWD_REVISION_A) ? 2 :
		       (revision == PCWD_REVISION_C) ? 4 : 8);
	unlock_kernel();
	printk(KERN_INFO "pcwd: Watchdog Module Unloaded.\n");

}

module_init(pcwatchdog_init);
module_exit(pcwatchdog_exit);

MODULE_AUTHOR("Holger Eiboeck");
MODULE_DESCRIPTION("Driver for the Berkshire PC watchdog cards");
MODULE_LICENSE("GPL");

EXPORT_NO_SYMBOLS;
